/**
  *  \file client/chart/markermode.cc
  *  \brief Starchart Mode: WMarkerChartMode
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chart/markermode.h"
#include "client/chart/deletedrawingmode.h"
#include "client/chart/movemarkermode.h"
#include "client/chart/drawmarkermode.h"
#include "client/chart/drawcirclemode.h"
#include "ui/event.h"
#include "util/nls.h"
#include "client/widgets/colorselect.h"
#include "client/chartloc.h"
#include "game/global.h"
#include "client/dialogs/helpdlg.h"
#include "ui/widgets/inputline.h"

// FIXME: observe current_turn, and automatically exit this mode
// when current_turn changes. Likewise, observe GDrawingContainer::sig_deleted
// and exit this mode when our drawing is deleted

// Since exiting this mode during processing will probably not work, we should
// first and foremost check for null whenever we access /drawing/.

/** Constructor. */
WMarkerChartMode::WMarkerChartMode(GDrawingContainer::iterator drawing)
    : WSimpleChartMode(ChartMode_PrimaryMode),
      drawing(drawing)
{ }

WMarkerChartMode::~WMarkerChartMode()
{ }

bool
WMarkerChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_ESCAPE:
         case vk_Quit:
            // Exit move mode
            delete this;
            return true;
         case 'c':
            // Change color
            if (*drawing == 0) {
                delete this;
            } else {
                WColorSelector csel;
                csel.setColor(GDrawing::getUserColorFromColor((*drawing)->getColor()));
                if (csel.doStandardDialog(_("Drawing Color"), false) == csel.rOK) {
                    (*drawing)->setColor(GDrawing::getColorFromUserColor(csel.getColor()));
                    loc.markChanged();
                }
            }
            return true;
         case SDLK_DELETE:
            // Delete marker
            if (*drawing != 0) {
                GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
                c.erase(drawing);
            }
            delete this;
            return true;
         case 'h':
         case ss_Alt + 'h':
            doHelp("pcc2:draw");
            return true;
        }
    }
    return WSimpleChartMode::handleEvent(event, loc);
}

/** Start moving a marker. Creates a WMoveMarkerChartMode and adds it to \c modes. */
void
WMarkerChartMode::moveMarker(WChartModeList& modes, GDrawingContainer::iterator p)
{
    modes.pushMode(*new WMoveMarkerChartMode(p));
}

/** Start drawing a line or rectangle. Creates a WDrawMarkerChartMode
    and adds it to \c modes. */
void
WMarkerChartMode::drawLineRectangle(WChartModeList& modes, GDrawingContainer::iterator p)
{
    modes.pushMode(*new WDrawMarkerChartMode(p));
}

/** Start drawing a circle. Creates a WDrawCircleChartMode and adds it to \c modes. */
void
WMarkerChartMode::drawCircle(WChartModeList& modes, GDrawingContainer::iterator p)
{
    modes.pushMode(*new WDrawCircleChartMode(p));
}

/** Start deleting a drawing. Creates a WDeleteDrawingChartMode and adds it to \c modes. */
void
WMarkerChartMode::deleteDrawing(WChartModeList& modes, GDrawingContainer::iterator p)
{
    modes.pushMode(*new WDeleteDrawingChartMode(p));
}


/*************************** editMarkerComment ***************************/

void
editMarkerComment(GDrawing& marker)
{
    string_t comment = marker.getComment();
    if (UIInputLine::run(comment, 255 /* FIXME? */, _("Marker Comment"), _("Enter new comment for this marker:"), 30, UIInputLine::ilf_GameChars))
        marker.setComment(strRTrim(comment));
}
