/**
  *  \file client/dialogs/labeldlg.cc
  *  \brief Ship and Planet Label Dialog
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/labeldlg.h"
#include "client/labels.h"
#include "client/lru.h"
#include "client/widgets/lrupopup.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "game/global.h"
#include "game/univ.h"
#include "gfx/palette.h"
#include "int/interror.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/unicodechars.h"
#include "client/widgets/helpwidget.h"
#include "ui/widgets/quit.h"

namespace {
    class WLabelExprEditDialog : public UIWindow {
        ObjectHolder h;
        UIInputLine* input[2];
        UIAbstractButton* button[2];

        void init();
        bool setLabel(const char* title, LabelKind kind);
     public:
        WLabelExprEditDialog();

        void onOK();
        void onCancel();
        void onPlanetDropdown();
        void onShipDropdown();
        void onDropdown(int what);

        bool handleEvent(const UIEvent& event, bool second_pass);
    };

    struct LRULink {
        LRUList* lru;
        LRUList* predef;
    };
}

CTASSERT(ShipLabel == 0);
CTASSERT(PlanetLabel == 1);

static const LRULink lru_links[] = {
    { &lru_shiplabels, &predef_shiplabels },
    { &lru_planetlabels, &predef_planetlabels },
};

WLabelExprEditDialog::WLabelExprEditDialog()
    : UIWindow(_("Ship and Planet Labels"), UIWindow::BLUE_WINDOW, COLOR_GRAY)
{
    init();
}

void
WLabelExprEditDialog::init()
{
    // VBox
    //   UIStaticText "Planets:"
    //   HBox
    //     UIInputLine
    //     UIButton dropdown
    //   UIStaticText "Ships:"
    //   HBox
    //     UIInputLine
    //     UIButton dropdown
    //  [UIMultilineStatic "blah blah"]
    //   UIHBox
    //     UIButton "OK"
    //     UIButton "Cancel"
    //     UISpacer
    //     UIButton "Help"

    setLayout(UIVBoxLayout::instance5);

    add(h.add(new UIStaticText(_("Planets:"), UIStaticText::StaticTitle)));

    UIGroup& g2 = h.add(new UIGroup(UIHBoxLayout::instance5));
    input[PlanetLabel] = &h.add(new UIInputLine(4096, 30, UIKey(ss_Alt + 'p'), UIInputLine::ilf_Framed, 1000));
    button[PlanetLabel] = &h.add(new UIButton(PlanetLabel, UTF_DOWN_ARROW, 0)).addCall(this, &WLabelExprEditDialog::onPlanetDropdown);
    g2.add(*input[PlanetLabel]);
    g2.add(*button[PlanetLabel]);
    add(g2);

    add(h.add(new UIStaticText(_("Ships:"), UIStaticText::StaticTitle)));

    UIGroup& g4 = h.add(new UIGroup(UIHBoxLayout::instance5));
    input[ShipLabel] = &h.add(new UIInputLine(4096, 30, UIKey(ss_Alt + 's'), UIInputLine::ilf_Framed, 1001));
    button[ShipLabel] = &h.add(new UIButton(ShipLabel, UTF_DOWN_ARROW, 0)).addCall(this, &WLabelExprEditDialog::onShipDropdown);
    g4.add(*input[ShipLabel]);
    g4.add(*button[ShipLabel]);
    add(g4);

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    add(g);

    g.add(h.add(new UIButton(2000, _("OK"),     SDLK_RETURN)).addCall(this, &WLabelExprEditDialog::onOK));
    g.add(h.add(new UIButton(2001, _("Cancel"), SDLK_ESCAPE)).addCall(this, &WLabelExprEditDialog::onCancel));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(_("Help"), 'h')));

    add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Tab)));
    add(h.add(new WHelpWidget("pcc2:labels")));
    add(h.add(new UIQuit(2001)));

    pack();

    input[PlanetLabel]->setText(getLabelExpr(PlanetLabel));
    input[ShipLabel]->setText(getLabelExpr(ShipLabel));
    input[PlanetLabel]->setFocusedWidget();
}

bool
WLabelExprEditDialog::setLabel(const char* title, LabelKind kind)
{
    try {
        setLabelExpr(kind, input[kind]->getText());
        return true;
    }
    catch (IntError& err) {
        messageBox(format(_("%s expression not accepted: %s"), title, err.what()), _("Ship and Planet Labels"));
        input[kind]->setFocusedWidget();
        getDisplayedTurn().getCurrentUniverse().sig_universe_changed.raise();
        return false;
    }
}

void
WLabelExprEditDialog::onOK()
{
    if (setLabel(_("Planet"), PlanetLabel) && setLabel(_("Ship"), ShipLabel))
        stop(1);
}

void
WLabelExprEditDialog::onCancel()
{
    stop(0);
}

void
WLabelExprEditDialog::onPlanetDropdown()
{
    onDropdown(PlanetLabel);
}

void
WLabelExprEditDialog::onShipDropdown()
{
    onDropdown(ShipLabel);
}
                                 
void
WLabelExprEditDialog::onDropdown(int slot)
{
    input[slot]->setFocusedWidget();
    const LRUListItem* item = WLRUPopup()
        .withList(_("Last expressions"),       *lru_links[slot].lru)
        .withList(_("Predefined expressions"), *lru_links[slot].predef)
        .doPopup(input[slot]->getText(), *button[slot]);

    if (item != 0) {
        input[slot]->setText(item->getValue());
    }
}

bool
WLabelExprEditDialog::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && event.key.code == SDLK_DOWN) {
        if (input[ShipLabel]->getFocusState() == PrimaryFocus) {
            onDropdown(ShipLabel);
            return true;
        }
        if (input[PlanetLabel]->getFocusState() == PrimaryFocus) {
            onDropdown(PlanetLabel);
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

/** Save label expressions into LRU lists. */
static void
saveLRU()
{
    for (int i = 0; i < 2; ++i) {
        string_t val = getLabelExpr(LabelKind(i));
        if (!val.empty() && lru_links[i].predef->findValue(val) == lru_links[i].predef->size()) {
            lru_links[i].lru->pushNewLimit(new LRUListItem(val, val));
        }
    }
}

void
editLabelExpressions()
{
    saveLRU();
    WLabelExprEditDialog dlg;
    ui_root->centerWidget(dlg);
    ui_root->runModal(dlg);
    saveLRU();
}
