/**
  *  \file game/planetpred.cc
  *  \brief Planet Prediction
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "game/planetpred.h"
#include "game/planet.h"
#include "game/host.h"
#include "game/formula.h"
#include "game/config.h"
#include "game/planetform.h"

GPlanetEffectors::GPlanetEffectors()
{
    clear();
}

void
GPlanetEffectors::clear()
{
    for (int i = 0; i < pe_MAX; ++i)
        effectors[i] = 0;
}

void
GPlanetEffectors::add(Effector eff, int count)
{
    effectors[eff] += count;
}

void
GPlanetEffectors::set(Effector eff, int count)
{
    effectors[eff] = count;
}

int
GPlanetEffectors::get(Effector eff) const
{
    return effectors[eff];
}

int
GPlanetEffectors::getNumTerraformers() const
{
    return effectors[pe_Heat50] + effectors[pe_Cool50] + effectors[pe_Heat100];
}


/****************************** computeTurn ******************************/

static void
trimHappiness(GPlanet& pl)
{
    if (pl.getColonistHappiness() < -300)
        pl.setColonistHappiness(mn16_t(-300));
    if (pl.getColonistHappiness() > 100)
        pl.setColonistHappiness(mn16_t(100));
    if (pl.getNativeHappiness() < -300)
        pl.setNativeHappiness(mn16_t(-300));
    if (pl.getNativeHappiness() > 100)
        pl.setNativeHappiness(mn16_t(100));
}

static void
doMining(GPlanet& pl, GCargoType el)
{
    int32_t cap = getMiningCapacity(pl, el, pl.getStructures(str_Mines));
    int32_t ore = pl.getOreGround(el);
    if (cap > ore)
        cap = ore;
    pl.changeCargoRaw(el, cap);
    pl.setOreGround(el, mp32_t(ore - cap));
}

static void
doTUDR(GPlanet& pl, GCargoType el)
{
    /* FIXME: this is the PHost formula. THost? */
    pl.setOreGround(el, mp32_t(pl.getOreGround(el) + (pl.getOreDensity(el) * config.TransuraniumDecayRate() + 50) / 100));
}

static void
doDecay(GPlanet& pl, TPlanetaryStructure str)
{
    int max = getMaxStructures(pl, str);
    int have = pl.getStructures(str);
    if (have > max) {
        have -= config.StructureDecayPerTurn(pl.getOwner());
        if (have < max)
            have = max;
        pl.setStructures(str, mp16_t(have));
    }
}

static void
doAssimilation(GPlanet& pl)
{
    if (raceId(pl.getOwner()) == 6 && (pl.getNativeRace() != 0 && pl.getNativeRace() != nr_Amorphous)) {
        int32_t assim = pl.getColonists() * config.BorgAssimilationRate(pl.getOwner()) / 100;
        int32_t natives = pl.getNatives();
        if (assim > natives)
            assim = natives;
        pl.changeCargoRaw(el_Colonists, assim);
        pl.setNatives(mp32_t(natives - assim));
    }
}

static void
doDemolish(GPlanet& pl, TPlanetaryStructure str, int n)
{
    int have = pl.getStructures(str);
    if (have > n)
        pl.setStructures(str, mp16_t(have - n));
    else
        pl.setStructures(str, mp16_t(0));
}

/** Compute a planet's turn.
    Based upon planacc.pas::ComputePlanetTurn.
    \param pl  Must be a <b>copy</b> of the actual planet
    \param eff Effectors for this planet*/
void
computeTurn(GPlanet& pl, const GPlanetEffectors& eff)
{
    /* Our sequence will be:
       - hiss
       - lfm / gather-build        (maybe)
       - dmp                       (maybe)
       - free fighters             (maybe)
       - terraform
       - mining
       - supplies
       - tudr
       - happiness
       - tax
       - THost: assimilate
       - growth
       - cdr
       - amorphs
       - riots
       - PHost: assimilate */

    /* Hiss */
    if (config.AllowHiss()) {
        int nhiss = eff.get(GPlanetEffectors::pe_Hiss);
        if (host.isPHost() && nhiss > config.MaxShipsHissing())
            nhiss = config.MaxShipsHissing();
        nhiss *= config.HissEffectRate(pl.getOwner());

        pl.setColonistHappiness(mn16_t(pl.getColonistHappiness() + nhiss));
        pl.setNativeHappiness  (mn16_t(pl.getNativeHappiness()   + nhiss));
        trimHappiness(pl);
    }

    /* LFM, Gather-build, free fighters would be here */

    /* Terraform */
    if (config.AllowScienceMissions()) {
        int rate = host.isPHost() ? config.TerraformRate(pl.getOwner()) : 1;
        int temp = pl.getTemperature();
        if (temp > 50) {
            /* Coolers */
            temp -= eff.get(GPlanetEffectors::pe_Cool50) * rate;
            if (temp < 50)
                temp = 50;
        }
        if (temp < 50) {
            /* Heaters */
            temp += eff.get(GPlanetEffectors::pe_Heat50) * rate;
            if (temp > 50)
                temp = 50;
        }
        /* Tholian heaters. Note that this is not 100% accurate. Given
           a 49F planet, one Bohemian (HeatTo50), one Onxy (HeatTo100).
           If the onyx goes first, we have 50F afterwards (because the
           bohemian sees a 50F planet and doesn't do anything). If the
           Bohemian goes first, we have 51F afterwards. */
        temp += eff.get(GPlanetEffectors::pe_Heat100) * rate;
        if (temp > 100)
            temp = 100;
        pl.setTemperature(mp16_t(temp));
    }

    /* FIXME: structure decay is before everything in THost */

    /* Mining */
    if (pl.getOwner() > 0 && pl.getOwner() <= NUM_PLAYERS) {
        doMining(pl, el_Neutronium);
        doMining(pl, el_Tritanium);
        doMining(pl, el_Duranium);
        doMining(pl, el_Molybdenum);

        /* Supplies */
        int32_t fact = pl.getStructures(str_Factories);
        if (host.isPHost())
            fact = fact * config.ProductionRate(pl.getOwner()) / 100;
        fact += getBovinoidSupplyContributionLimited(pl);
        pl.changeCargoRaw(el_Supplies, fact);
    }

    /* TUDR */
    doTUDR(pl, el_Neutronium);
    doTUDR(pl, el_Tritanium);
    doTUDR(pl, el_Duranium);
    doTUDR(pl, el_Molybdenum);

    /* Happiness */
    if (pl.getOwner() > 0 && pl.getOwner() <= NUM_PLAYERS) {
        if (pl.getColonistHappiness() < 30)
            pl.setColonistTax(mp16_t(0));
        if (pl.getNativeHappiness() < 30)
            pl.setNativeTax(mp16_t(0));
        pl.setColonistHappiness(mn16_t(pl.getColonistHappiness() + getColonistChange(pl)));
        pl.setNativeHappiness(mn16_t(pl.getNativeHappiness() + getNativeChange(pl)));
        trimHappiness(pl);

        /* Taxation */
        int32_t limit = host.getPostTaxationHappinessLimit();
        int32_t income = 0;
        if (pl.getColonistHappiness() >= limit) {
            income += getColonistDue(pl, pl.getColonistTax());
        }
        if (pl.getNativeHappiness() >= limit) {
            income += getNativeDueLimited(pl, pl.getNativeTax(), 0x7FFFFFFF);
        }
        if (income > config.MaxPlanetaryIncome(pl.getOwner())) {
            income = config.MaxPlanetaryIncome(pl.getOwner());
        }
        pl.changeCargoRaw(el_Money, income);

        /* PHost: Assimilation */
        /* FIXME: seems to be here as well in HOST */
        if (host.isPHost())
            doAssimilation(pl);

        /* Colonist growth */
        limit = getMaxSupportedColonists(pl);
        if (host.isPHost()) {
            /* PHost 4.1/3.5 changed rounding in mining, as well as in overpopulation
               deaths; the latter implicitly due to the computation of deaths in
               persons, not clans. */
            const int ceilOrTrunc = host.isPHostRoundingMiningResults() ? 0 : 99;
            if (pl.getColonistHappiness() >= 70 && pl.getColonists() < limit) {
                double rate;
                if (raceId(pl.getOwner()) == 7 && config.CrystalsPreferDeserts()) {
                    if (config.CrystalSinTempBehavior()) {
                        if (pl.getTemperature() < 15) {
                            rate = 0.0;
                        } else {
                            rate = 5.0 * std::sin(pl.getTemperature() * PI/200);
                        }
                    } else {
                        rate = (5*pl.getTemperature()) / 100.0;
                    }
                } else {
                    if (pl.getTemperature() < 15 || pl.getTemperature() > 84) {
                        rate = 0.0;
                    } else {
                        rate = 5.0 * std::sin(pl.getTemperature() * PI/100);
                    }
                }

                /* Note this rounding behaviour is right */
                int32_t growth = roundToInt(rate * pl.getColonists() / ((1 + pl.getColonistTax()/5.0) * 100))
                    * config.RaceGrowthRate(pl.getOwner()) / 100;
                int32_t maxGrowth = limit - pl.getColonists();
                if (growth > maxGrowth)
                    growth = maxGrowth;
                pl.changeCargoRaw(el_Colonists, growth);
            }

            if (pl.getColonists() > limit && config.ClimateLimitsPopulation()) {
                int32_t deaths = (config.ClimateDeathRate(pl.getOwner()) * pl.getColonists() + ceilOrTrunc) / 100;
                int32_t maxDeaths = pl.getColonists() - limit;
                if (deaths > maxDeaths)
                    deaths = maxDeaths;
                pl.changeCargoRaw(el_Colonists, -deaths);
            }

            /* Native growth */
            if (pl.getNativeRace()) {
                bool likeItHot = (pl.getNativeRace() == nr_Siliconoid
                                  && config.CrystalsPreferDeserts()
                                  && host.hasSiliconoidDesertAdvantage());
                if (!config.ClimateLimitsPopulation()) {
                    limit = 156000;
                } else if (likeItHot) {
                    limit = pl.getTemperature() * 1000L;
                } else {
                    limit = int32_t(std::sin(pl.getTemperature() * PI / 100) * 156000);
                }

                if (pl.getNativeHappiness() >= 70 && pl.getNatives() < limit) {
                    int32_t growth;
                    if (likeItHot) {
                        growth = (pl.getTemperature() * pl.getNatives()) / (500 * (pl.getNativeTax() + 5));
                    } else {
                        growth = int32_t((4*std::sin(pl.getTemperature() * PI/100) / (1 + pl.getNativeTax()/5.0)) * pl.getNatives()) / 100;
                    }
                    int32_t nn = pl.getNatives() + growth;
                    if (nn > limit)
                        nn = limit;
                    pl.setNatives(mp32_t(nn));
                }

                if (pl.getNatives() > limit && config.ClimateLimitsPopulation()) {
                    int32_t deaths = (config.NativeClimateDeathRate() * pl.getNatives() + ceilOrTrunc) / 100;
                    int32_t nn = pl.getNatives() - deaths;
                    if (nn < limit)
                        nn = limit;
                    pl.setNatives(mp32_t(nn));
                }
            }

            /* FIXME: Supply eat */
        } else {
            /* THost */
            int32_t growth = 0;
            if (pl.getColonistHappiness() >= 70) {
                // These formulas have the form
                //     (temperature term) * colonists/20 * 5/(5+tax)
                // which simplifies to
                //     (temperature term) * colonists / (4 * (5+tax))
                if (raceId(pl.getOwner()) == 7) {
                    growth = radiv(pl.getTemperature() * pl.getColonists(), 100 * 4 * (pl.getColonistTax()+5));
                } else {
                    growth = roundToInt((std::sin((100 - pl.getTemperature()) * 0.0314) * pl.getColonists()) / (4 * (pl.getColonistTax() + 5.0)));
                }
            }

            int32_t maxPop = roundToInt(std::sin((100 - pl.getTemperature()) * 0.0314) * 100000);
            if (pl.getTemperature() <= 14 || pl.getTemperature() > 84) {
                if (raceId(pl.getOwner()) != 7 || !config.CrystalsPreferDeserts()) {
                    growth = 0;
                    maxPop = std::max(1, pl.getColonists() - (pl.getColonists() * config.ClimateDeathRate(pl.getOwner()) / 100));
                }
                if (pl.getTemperature() > 50) {
                    maxPop += (100-pl.getTemperature()) * 2;
                } else {
                    maxPop += (1+pl.getTemperature())*2;
                }
            }
            if (raceId(pl.getOwner()) == 7 && config.CrystalsPreferDeserts()) {
                maxPop = pl.getTemperature() * 1000;
            }
            if (pl.getTemperature() > 80 && (raceId(pl.getOwner()) >= 9 || raceId(pl.getOwner()) == 4)) {
                maxPop = std::max(maxPop, 60);
            }
            if (pl.getTemperature() < 20 && raceId(pl.getOwner()) == 10) {
                maxPop = 90000;
            }
            if (!config.ClimateLimitsPopulation()) {
                maxPop = 100000;
            }

            if (config.AllowEatingSupplies(pl.getOwner()) && pl.getColonists() > maxPop) {
                int32_t eaten = rdivaddl(pl.getColonists() - maxPop, 40, 1);
                int32_t maxEaten = pl.getCargo(el_Supplies);
                if (eaten > maxEaten)
                    eaten = maxEaten;
                pl.changeCargoRaw(el_Supplies, -eaten);
                maxPop = rdivaddl(pl.getCargo(el_Supplies), 40, maxPop);
            }

            if (pl.getColonists() > 66000) {
                growth = rdivaddl(growth, 2, 0);
            }

            if (pl.getTemperature() > 84 || pl.getTemperature() <= 14) {
                if (raceId(pl.getOwner()) == 7 && pl.getTemperature() > 50) {
                    pl.changeCargoRaw(el_Colonists, growth);
                }
            } else if (pl.getColonists() < maxPop) {
                pl.changeCargoRaw(el_Colonists, growth);
            }
            if (pl.getColonists() > maxPop) {
                pl.setColonists(mn32_t(maxPop));
            }
            if (pl.getColonists() > 250000) {
                pl.setColonists(mn32_t(250000));
            }

            /* Native Growth */
            if (pl.getNativeRace() != 0) {
                bool siliconoid = (pl.getNativeRace() == nr_Siliconoid);
                if (pl.getNativeHappiness() < 70) {
                    growth = 0;
                } else if (siliconoid) {
                    growth = rdivaddl(pl.getTemperature() * pl.getNatives() * 5, 100 * 25 * (pl.getNativeTax() + 5), 0);
                } else {
                    growth = roundToInt(std::sin((100-pl.getTemperature()) * 0.0314) * pl.getNatives() * 5
                                        / (25 * (pl.getNativeTax() + 5)));
                }

                if (siliconoid) {
                    maxPop = 1000 * pl.getTemperature();
                } else {
                    maxPop = roundToInt(std::sin((100-pl.getTemperature()) * 0.0314) * 150000);
                }

                if (pl.getNatives() > 66000) {
                    growth = rdivaddl(growth, 2, 0);
                }
                if (pl.getNatives() < maxPop) {
                    pl.setNatives(mp32_t(pl.getNatives() + growth));
                }
            }
        }

        /* Structure decay */
        doDecay(pl, str_Mines);
        doDecay(pl, str_Factories);
        doDecay(pl, str_Defense);

        /* Riots */
        if (!host.isPHost()) {
            /* FIXME */
        } else {
            if ((pl.getColonists() > 0 && pl.getColonistHappiness() < 40)
                || (pl.getNatives() > 0 && pl.getNativeHappiness() < 40))
            {
                doDemolish(pl, str_Factories, 10);
                doDemolish(pl, str_Mines, 10);
            }

            int32_t colDeaths = (pl.getColonistHappiness() < 20
                                 ? pl.getColonists() * (40 - pl.getColonistHappiness()) / 5
                                 : pl.getNatives() > 0 && pl.getNativeHappiness() < 20
                                 ? pl.getColonists() * (40 - pl.getNativeHappiness()) / 25
                                 : 0);
            int32_t natDeaths = (pl.getNativeHappiness() < 20
                                 ? pl.getNatives() * (40 - pl.getNativeHappiness()) / 5
                                 : pl.getColonists() > 0 && pl.getColonistHappiness() < 20
                                 ? pl.getNatives() * (40 - pl.getColonistHappiness()) / 25
                                 : 0);

            /* Colonist deaths */
            colDeaths = (colDeaths+99) / 100;
            int32_t maxColDeaths = pl.getColonists();
            if (colDeaths > maxColDeaths)
                colDeaths = maxColDeaths;
            pl.changeCargoRaw(el_Colonists, -colDeaths);

            /* Native deaths */
            natDeaths = (natDeaths+99) / 100;
            int32_t natives = pl.getNatives() - natDeaths;
            if (natives < 0)
                natives = 0;
            pl.setNatives(mp32_t(natives));
        }

        /* Amorphs */
        if (pl.getNatives() != 0 && pl.getNativeRace() == nr_Amorphous) {
            int32_t eaten = getAmorphousBreakfast(pl.getNativeHappiness());
            int32_t maxEaten = pl.getColonists();
            if (eaten > maxEaten)
                eaten = maxEaten;
            pl.changeCargoRaw(el_Colonists, -eaten);
        }

        /* THost assim */
        if (!host.isPHost())
            doAssimilation(pl);
    }

    /* Clean up */
    if (pl.getNativeRace() == 0 || pl.getNatives() == 0) {
        pl.setNativeRace(mp16_t(0));
        pl.setNatives(mp32_t(0));
    }
    if (pl.getColonists() == 0 || pl.getOwner() == 0) {
        pl.setColonists(mn32_t(0));
        // FIXME: we cannot yet do that: pl.setOwner(mp16_t(0));
    }
}
