/**
  *  \file util/helpindex.cc
  *  \brief Class HelpIndex
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <stdexcept>
#include <algorithm>
#include "util/helpindex.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "util/xmlread.h"
#include "util/misc.h"

namespace {
    struct CompareNodes {
        bool operator()(const HelpIndex::Node* pa, const HelpIndex::Node* pb) const
            {
                // Sort by priority (lower value goes first),
                //    -- this is the order in which merged pages appear
                // then age (higher value goes first),
                //    -- this means later files replace older ones
                // then position (lower value goes first).
                //    -- error recovery: if a file contains multiple pages of same value, only use first one
                if (pa->priority != pb->priority) {
                    return pa->priority < pb->priority;
                }
                if (pa->file.serial != pb->file.serial) {
                    return pa->file.serial > pb->file.serial;
                }
                return pa->pos < pb->pos;
            }
    };
}

HelpIndex HelpIndex::instance;

HelpIndex::HelpIndex()
    : m_files(),
      m_nodes(),
      m_counter(0)
{ }

HelpIndex::~HelpIndex()
{ }

void
HelpIndex::addFile(string_t name, string_t origin)
{
    File* p = m_files.push_back(new File());
    p->name = name;
    p->origin = origin;
    p->scanned = false;
    p->serial = m_counter++;
}

void
HelpIndex::removeFilesByOrigin(string_t origin)
{
    // Remove all nodes that point at a file of this origin
    for (std::multimap<string_t, Node>::iterator it = m_nodes.begin(); it != m_nodes.end(); ) {
        if (it->second.file.origin == origin) {
            m_nodes.erase(it++);
        } else {
            ++it;
        }
    }

    // Remove all files of this origin
    for (size_t i = 0; i < m_files.size(); ) {
        if (m_files[i]->origin == origin) {
            m_files.erase(m_files.begin() + i);
        } else {
            ++i;
        }
    }
}

void
HelpIndex::find(string_t page, std::vector<Node*>& out)
{
    // Make sure index is up-to-date
    scanNewFiles();

    // Build list of possible nodes
    std::multimap<string_t, Node>::iterator it = m_nodes.lower_bound(page);
    while (it != m_nodes.end() && it->first == page) {
        out.push_back(&it->second);
        ++it;
    }

    // Sort and filter out duplicates
    // Note that "toc" is never filtered
    std::sort(out.begin(), out.end(), CompareNodes());
    if (!out.empty() && page != "toc") {
        std::vector<Node*>::iterator it = out.begin();
        int pri = (*it)->priority;
        ++it;

        while (it != out.end()) {
            if (pri == (*it)->priority) {
                it = out.erase(it);
            } else {
                pri = (*it)->priority;
                ++it;
            }
        }
    }
}

void
HelpIndex::scanNewFiles()
{
    for (size_t i = 0, n = m_files.size(); i < n; ++i) {
        if (!m_files[i]->scanned) {
            m_files[i]->scanned = true;
            try {
                scanFile(*m_files[i]);
                console.write(LOG_INFO, format(_("Scanned help file %s."), m_files[i]->name));
            }
            catch (std::exception& e) {
                console.write(LOG_ERROR, format("%s: %s", m_files[i]->name, e.what()));
            }
        }
    }
}

void
HelpIndex::scanFile(const File& file)
{
    Ptr<Stream> s = Stream::openFile(file.name, Stream::C_READ);
    XmlReader rdr(*s);

    std::vector<int> priorityStack;
    priorityStack.push_back(100);

    bool atPageStart = false;
    bool atGroupStart = false;
    Stream::fpos_t pagePos = 0;
    while (1) {
        switch (rdr.read()) {
         case XmlReader::TagStart:
            atPageStart = (rdr.getTag() == "page");
            atGroupStart = (rdr.getTag() == "group" || rdr.getTag() == "help");
            pagePos = rdr.getPosition();
            priorityStack.push_back(priorityStack.back());
            break;
         case XmlReader::TagAttribute:
            if (atPageStart && rdr.getName() == "id") {
                m_nodes.insert(std::make_pair(rdr.getValue(), Node(priorityStack.back(), file, pagePos)));
            }
            if (atGroupStart && rdr.getName() == "priority") {
                stringValue(rdr.getValue(), priorityStack.back());
            }
            break;
         case XmlReader::TagEnd:
            if (priorityStack.size() > 1) {
                priorityStack.pop_back();
            }
            break;
         case XmlReader::Eof:
            goto out;
         default:
            break;
        }
    }
 out:;
}
